import java.awt.Color;
import java.awt.FontMetrics;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Rectangle;
import java.awt.Shape;
import java.awt.geom.Path2D;
import java.awt.geom.Rectangle2D;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.swing.JPanel;

/**
 * Plot datasets
 * 
 * 
 * @version 1.1 March 2010
 * @author Stephan Euler
 * 
 */
public class Plotter extends JPanel {

	/**
	 * 
	 */
	private static final long serialVersionUID = 8220278066243017406L;

	static int verbose = 0;
	static Color[] plotColor = { Color.red, Color.green, Color.blue,
			Color.orange, Color.yellow, Color.magenta };

	Map<String, DataObject> dataObjects = new HashMap<String, DataObject>();
	String currV = "0";

	private double xmin = -1;
	private double xmax = 1;
	private double ymin = -1;
	private double ymax = 1;
	private double[] xgrid = null;
	private double[] ygrid = null;
	private String[] xlabel = null;
	private List<TextObject> textObjects = new ArrayList<TextObject>();
	private boolean dynYscale = true;
	private boolean dynXscale = true;
	private double xl = 0.05;
	private double xr = 0.95;
	private double yu = 0.05;
	private double yl = 0.95;
	private List<Double> yLine = new ArrayList<Double>();
	private List<Double> xLine = new ArrayList<Double>();

	private double autoXgrid = 0; // distance between xtics (sec)
	private double ticRelHeight = 0.05;
	private int symbolSize = 2;
	private double autoYgrid = 0;
	private LineStyle plotMode = LineStyle.LINE;
	private int plotLeft;
	private double xfactor;
	private double yfactor;
	private int plotLow;

	private double halfBarWidth = 0.4;
	private String yLabelFormat = null;
	private String xLabelFormat = null;

	/**
	 * Constructs a new <code>Plotter</code>
	 * 
	 */
	public Plotter() {
		this("Plotter");
	}

	/**
	 * Constructs a new <code>Plotter</code> with the given name
	 * 
	 * @param name
	 *            the name of the new <code>Plotter</code> object
	 */
	public Plotter(String name) {
		setName(name);
		setBackground(Color.lightGray);
	}

	/**
	 * remove all data sets
	 */
	public void clearPlotVector() {
		dataObjects.clear();
	}

	/**
	 * Add a point to the current data set
	 * 
	 * @param x
	 *            x coordinate
	 * @param y
	 *            y coordinate
	 */
	public void add(double x, double y) {
		add(currV, x, y);
	}

	/**
	 * Add a point to the specified data set
	 * 
	 * @param key
	 *            key for the data set
	 * @param x
	 *            x coordinate
	 * @param y
	 *            y coordinate
	 */
	public synchronized void add(String key, double x, double y) {
		if (dataObjects.containsKey(key)) {
			DataObject d = dataObjects.get(key);
			double[] pv = d.getData();
			double[] tmp = new double[pv.length + 2];
			System.arraycopy(pv, 0, tmp, 0, pv.length);
			tmp[tmp.length - 2] = x;
			tmp[tmp.length - 1] = y;
			d.setData(tmp);
			// dataSets.put(key, tmp);
			// System.out.println( Arrays.toString(tmp));
		} else {
			DataObject d = new DataObject();
			d.setColor(plotColor[dataObjects.size() % plotColor.length]);
			double[] pv = new double[2];
			pv[0] = x;
			pv[1] = y;
			d.setData(pv);
			dataObjects.put(key, d);
		}
	}

	/**
	 * Set the color for the specified data set.
	 * 
	 * @param key
	 *            key for the data set
	 * @param c
	 *            color
	 */
	public void setDataColor(String key, Color c) {
		if (!dataObjects.containsKey(key)) {
			DataObject d = new DataObject();
			dataObjects.put(key, d);
		}
		dataObjects.get(key).setColor(c);
	}

	/**
	 * Add a point to the current data set, use index as x coordinate
	 * 
	 * @param y
	 *            y coordinate
	 */
	public void add(double y) {
		if (dataObjects.containsKey(currV)) {
			double[] pv = dataObjects.get(currV).getData();
			add(pv.length / 2, y);
		} else {
			add(0, y);
		}
	}

	/**
	 * Write text at the specified position
	 * 
	 * @param s
	 *            text to write
	 * @param x
	 *            x coordinate
	 * @param y
	 *            y coordinate
	 */
	public TextObject setText(String s, double x, double y) {
		TextObject t = new TextObject(s, x, y);
		textObjects.add(t);
		return t;
	}

	/**
	 * Draw a horizontal line
	 * 
	 * @param y
	 *            where to draw the line
	 */
	public void setYLine(double y) {
		yLine.add(y);
	}

	/**
	 * Draw a vertical line
	 * 
	 * @param x
	 *            where to draw the line
	 */
	public void setXLine(double x) {
		xLine.add(x);
	}

	/**
	 * Set values for drawing x ticks with default labels.
	 * 
	 * @param x
	 */
	public void setXGrid(double[] x) {
		xgrid = new double[x.length];
		xlabel = new String[x.length];
		for (int i = 0; i < x.length; i++) {
			xgrid[i] = x[i];
			xlabel[i] = getXLabel(x[i]);
		}
		// yl = 0.9; // make more space for labels
	}

	private String getXLabel(double d) {
		if (xLabelFormat == null)
			return "" + d;
		else
			return String.format(xLabelFormat, d);
	}

	/**
	 * Define labels for the x ticks.
	 * 
	 * @param label
	 */
	public void setXLabel(String[] label) {
		xlabel = new String[label.length];
		for (int i = 0; i < label.length; i++) {
			xlabel[i] = label[i];
		}
	}

	/**
	 * 
	 */
	public void setYGrid(double[] y) {
		ygrid = new double[y.length];
		for (int i = 0; i < y.length; i++) {
			ygrid[i] = y[i];
		}
		xl = 0.1; // make more space for labels
		yu = 0.1;
	}

	/**
	 * Set the x range for plotting
	 * 
	 * @param min
	 *            Minimum value to plot
	 * @param max
	 *            Maximuma value to plot
	 */
	public void setXrange(double min, double max) {
		xmin = min;
		xmax = max;
		dynXscale = false;
	}

	/**
	 * Set the y range for plotting
	 * 
	 * @param min
	 *            Minimum value to plot
	 * @param max
	 *            Maximuma value to plot
	 */
	public void setYrange(double min, double max) {
		ymin = min;
		ymax = max;
		dynYscale = false;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see javax.swing.JComponent#paint(java.awt.Graphics)
	 */
	public synchronized void paint(Graphics g) {
		Graphics2D g2 = (Graphics2D) g;
		if (verbose > 0)
			System.out.println("PAINT");

		int rectWidth = getSize().width - getInsets().left - getInsets().right;
		int rectHeight = getSize().height - getInsets().top
				- getInsets().bottom;
		// System.out.println("x: "
		// + getSize().width+" "+getInsets().left+" "+getInsets().right);
		// System.out.println("y: "
		// + getSize().height+" "+getInsets().top+" "+getInsets().bottom);

		plotLeft = (int) (rectWidth * xl) + getInsets().left;
		int plotRight = (int) (rectWidth * xr) + getInsets().left;
		int plotWidth = plotRight - plotLeft;
		int plotUp = (int) (rectHeight * yu) + getInsets().top;
		plotLow = (int) (rectHeight * yl) + getInsets().top;
		int plotHeight = plotLow - plotUp;

		double yw;
		if (dynYscale && dataObjects.size() > 0) {
			ymax = Double.MIN_VALUE;
			ymin = Double.MAX_VALUE;
			for (DataObject d : dataObjects.values()) {
				double[] vek = d.getData();
				ymax = Vis.vmax(ymax, vek, 1, 2);
				ymin = Vis.vmin(ymin, vek, 1, 2);
			}
			yw = ymax - ymin;
			ymin -= 0.125 * yw;
			yw *= 1.25;
			ymax = ymin + yw;

		} else {
			yw = (ymax - ymin);
		}
		yfactor = plotHeight / yw;

		double xw = (xmax - xmin);
		if (dynXscale && dataObjects.size() > 0) {
			xmax = Double.MIN_VALUE;
			xmin = Double.MAX_VALUE;
			for (DataObject d : dataObjects.values()) {
				double[] vek = d.getData();
				xmax = Vis.vmax(xmax, vek, 0, 2);
				xmin = Vis.vmin(xmin, vek, 0, 2);
			}
			xw = xmax - xmin;
		}
		xfactor = plotWidth / xw;

		g2.setColor(java.awt.Color.blue);
		for (double yL : yLine) {
			int yg = plotLow - (int) ((yL - ymin) * yfactor);
			g2.drawLine(plotLeft, yg, plotRight, yg);
		}

		for (double xL : xLine) {
			int xg = (int) (plotLeft + (xL - xmin) * xfactor);
			// System.out.println( "xLine: " + xL + " " + xg);
			g2.drawLine(xg, plotLow, xg, plotUp);
		}

		g2.setColor(java.awt.Color.blue);
		g2.drawRect(plotLeft, plotUp, plotWidth, plotHeight);

		for (TextObject tO : textObjects) {
			FontMetrics fm = g2.getFontMetrics();
			int xlabpos = scaleX(tO.getX()) - fm.stringWidth(tO.getText()) / 2;
			int ylabpos = scaleY(tO.getY()) + fm.getHeight() / 2;

			g2.setColor(tO.getColor());
			g2.drawString(tO.getText(), xlabpos, ylabpos);
		}

		if (autoXgrid > 0) {
			for (double i = xmin; i < xmax; i += autoXgrid) {
				int xg = plotLeft + (int) ((i - xmin) * xfactor);
				g2.drawLine(xg, plotLow, xg, plotLow
						- (int) (ticRelHeight * plotHeight));
				FontMetrics fm = g2.getFontMetrics();
				String label;
				if (xLabelFormat != null) {
					label = String.format(xLabelFormat, i);
				} else {
					label = "" + i;
				}
				int xlabpos = xg - fm.stringWidth(label) / 2;
				g2.drawString(label, xlabpos, getSize().height
						- getInsets().bottom);
			}
		}

		if (autoYgrid > 0) {

			// first get all y labels
			List<Double> yValues = new ArrayList<Double>();
			if (ymin * ymax < 0) {
				for (double yL = 0; yL <= ymax; yL += autoYgrid) {
					yValues.add(yL);
				}
				for (double yL = 0; yL >= ymin; yL -= autoYgrid) {
					yValues.add(yL);
				}

			} else {
				double ystart =  ymin;
				if( Math.IEEEremainder(ystart, autoYgrid) < 1.e-10 ) {
					int m = (int) (ystart / autoYgrid );
					ystart = (m+1)* autoYgrid;
				}
				for (double yL = ystart; yL <= ymax; yL += autoYgrid) {
					yValues.add(yL);
				}
			}

			// now draw them
			for (double yL : yValues) {
				int yg = scaleY(yL); // plotLow - (int) ((yL - ymin) * yfactor);
				g2.drawLine(plotLeft, yg, plotLeft
						+ (int) (ticRelHeight * plotWidth), yg);
				g2.drawString(getYLabel(yL), getInsets().left, yg);
			}
		}

		if (xgrid != null) {
			FontMetrics fm = g2.getFontMetrics();
			for (int i = 0; i < xgrid.length; i++) {
				int xg = plotLeft + (int) ((xgrid[i] - xmin) * xfactor);
				g2.drawLine(xg, plotLow, xg, plotLow - (int) (ticRelHeight * plotHeight));
				if (xlabel != null) {
					int xlabpos = xg - fm.stringWidth(xlabel[i]) / 2;
					g2.drawString(xlabel[i], xlabpos, getSize().height
							- getInsets().bottom);
				}
			}
		}

		if (ygrid != null) {
			for (int i = 0; i < ygrid.length; i++) {
				if (ygrid[i] >= ymin && ygrid[i] <= ymax) {
					int yg = plotLow - (int) ((ygrid[i] - ymin) * yfactor);
					g2.drawLine(plotLeft, yg, plotRight, yg);
					g2.drawString("" + ygrid[i], getInsets().left, yg);
				}
			}
		}

		// if (dataObjects.size() > 0) {
		int xcol = 0;
		for (DataObject d : dataObjects.values()) {
			LineStyle currStyle = plotMode;
			if (d.getLineStyle() != LineStyle.UNDEFINED)
				currStyle = d.getLineStyle();
			double[] tmpXVector = d.getData();
			if (verbose > 0)
				System.out.println("plotting  points");

			Path2D.Double pd = new Path2D.Double();
			g2.setColor(d.getColor());

			for (int i = 0; i < tmpXVector.length; i += 2) {
				double x = scaleX(tmpXVector[i]);
				double y = scaleY(tmpXVector[i + 1]);

				if (currStyle == LineStyle.LINE | currStyle == LineStyle.BOTH) {
					if (i == 0) {
						pd.moveTo(x, y);
					} else {
						pd.lineTo(x, y);
					}
				}

				double yref = 0;
				if (ymin > 0)
					yref = ymin;

				if (currStyle == LineStyle.IMPULS) {
					g2.drawLine((int) x, scaleY(yref), (int) x, (int) y);
				}

				if (currStyle == LineStyle.HISTOGRAM) {
					int width = scaleX(tmpXVector[i] + halfBarWidth)
							- scaleX(tmpXVector[i] - halfBarWidth);
					int height = scaleY(yref) - (int) y;
					Shape s = new Rectangle2D.Double(scaleX(tmpXVector[i]
							- halfBarWidth), (int) y, width, height);
					g2.draw(s);
					g2.fill(s);
				}

				if (currStyle == LineStyle.SYMBOL | currStyle == LineStyle.BOTH) {
					g2.drawOval((int) (x - symbolSize / 2),
							(int) (y - symbolSize / 2), symbolSize, symbolSize);
				}
				if (currStyle == LineStyle.DOT) {
					g2.drawOval((int) x, (int) y, 1, 1);
				}

				if (currStyle == LineStyle.LINE | currStyle == LineStyle.BOTH) {
					// Stroke drawingStroke3 = new BasicStroke(3,
					// BasicStroke.CAP_BUTT, BasicStroke.JOIN_ROUND, 0,
					// new float[] { 5 }, 0);
					g2.draw(pd);

				}
				++xcol;
			}
		}
	}

	private String getYLabel(double yl2) {
		if (yLabelFormat == null)
			return "" + yl2;
		else
			return String.format(yLabelFormat, yl2);
	}

	private int scaleY(double d) {
		return (int) (plotLow - (d - ymin) * yfactor);
	}

	private int scaleX(double d) {
		return (int) (plotLeft + (d - xmin) * xfactor);
	}

	/**
	 * Switch to next data set.
	 * 
	 * This creates a new empty data set. The new set is default target for
	 * subsequent calls of the methode <code>add</code>.
	 */
	public void nextVector() {
		int i = 1;
		do {
			currV = "" + i;
			++i;
		} while (dataObjects.containsKey(currV));
		System.out.println("next vector: " + currV);
	}

	/**
	 * Set the spacing of labels on the x axis.
	 * 
	 * @param autoXgrid
	 *            the autoXgrid to set
	 */
	public void setAutoXgrid(double autoXgrid) {
		this.autoXgrid = autoXgrid;
	}

	/**
	 * Set the spacing of labels on the y axis.
	 * 
	 * @param autoYgrid
	 *            the autoYgrid to set
	 */
	public void setAutoYgrid(double autoYgrid) {
		this.autoYgrid = autoYgrid;
	}

	/**
	 * Get the current plot mode.
	 * 
	 * @return the plotMode
	 */
	public LineStyle getPlotMode() {
		return plotMode;
	}

	/**
	 * Set the plot mode. Currently available:
	 * <ul>
	 * <li>LINE: data points are connected with lines</li>
	 * <li>SYMBOL: draw a symbol at each data point</li>
	 * <li>BOTH: combination of LINE and SYMBOL</li>
	 * <li>IMPULS: verticle line to point</li>
	 * <li>HISTOGRAM: verticle box to pointL</li>
	 * </ul>
	 * 
	 * @param lineStyle
	 *            the plotMode to set
	 */
	public void setDataLineStyle(LineStyle lineStyle) {
		setDataLineStyle(currV, lineStyle);
		plotMode = lineStyle;
	}

	/**
	 * Set the plot mode.
	 * 
	 * @see setDataLineStyle
	 * 
	 * @param key
	 * @param lineStyle
	 */
	public void setDataLineStyle(String key, LineStyle lineStyle) {
		checkForObject(key);
		dataObjects.get(key).setLineStyle(lineStyle);
	}

	private void checkForObject(String key) {
		if (!dataObjects.containsKey(key)) {
			DataObject d = new DataObject();
			dataObjects.put(key, d);
		}
	}

	/**
	 * @return the symbolSize
	 */
	public int getSymbolSize() {
		return symbolSize;
	}

	/**
	 * Set the size of the symbols used in some plotting styles
	 * 
	 * @param symbolSize
	 *            the symbolSize to set
	 */
	public void setSymbolSize(int symbolSize) {
		this.symbolSize = symbolSize;
	}

	/**
	 * Returns a string with status information
	 * 
	 * @return the string
	 */
	public String getStatusLine() {
		String status = "";
		status += dataObjects.size() + " Sets, ";
		if (textObjects.size() == 1) {
			status += " 1 text object, ";
		}
		if (textObjects.size() > 1) {
			status += textObjects.size() + " text objects, ";
		}
		status += "x: [" + String.format("%.3g", xmin) + ","
				+ String.format("%.3g", xmax) + "] ";
		status += "y: [" + String.format("%.3g", ymin) + ","
				+ String.format("%.3g", ymax) + "] ";
		return status;
	}

	/**
	 * Remove the first text object that matches the given string. Returns
	 * <code>true</code> if this plotter contained the specified element.
	 * 
	 * @param string
	 *            element to be removed from this plotter, if present
	 * @return <code>true</code> if this plotter contained the specified element
	 */
	public boolean removeText(String string) {
		for (TextObject tO : textObjects) {
			if (string == tO.getText()) {
				textObjects.remove(tO);
				return true;
			}
		}
		return false;
	}

	/**
	 * Get the value of half the width of the boxes in plot mode HISTOGRAM
	 * 
	 * @return the halfBarWidth
	 */
	public double getHalfBarWidth() {
		return halfBarWidth;
	}

	/**
	 * Set the half the width of the boxes in plot mode HISTOGRAM
	 * 
	 * @param halfBarWidth
	 *            the halfBarWidth to set
	 */
	public void setHalfBarWidth(double halfBarWidth) {
		this.halfBarWidth = halfBarWidth;
	}

	/**
	 * Define the formatting of the labels of the y-axis. The given string will
	 * be used in <code>String.format(labelFormat, y)</code> to format a value
	 * <code>y</code>. For example <code>"%f.2%%"</code> will print numbers with
	 * two decimal place and a %-symbol. Default is <code>null</code>, i. e. the
	 * label is converted directly into a String.
	 * 
	 * @param labelFormat
	 *            the yLabelFormat to set
	 */
	public void setYLabelFormat(String labelFormat) {
		yLabelFormat = labelFormat;
	}

	/**
	 * Define the formatting of the labels of the x-axis. The given string will
	 * be used in <code>String.format(labelFormat, x)</code> to format a value
	 * <code>x</code>. For example <code>"%f.2%%"</code> will print numbers with
	 * two decimal place and a %-symbol. Default is <code>null</code>, i. e. the
	 * label is converted directly into a String.
	 * 
	 * @param labelFormat
	 *            the xLabelFormat to set
	 */
	public void setXLabelFormat(String labelFormat) {
		xLabelFormat = labelFormat;
	}
}
